/*------------------------------------------------------------------------
  File        : rtb_events.p
  Description : Roundtable published event handler
  Notes       :
  
    The procedure is run persistently in the Roundtable session and
    receives the published event as documented below. To intercept an
    invent, modify 'evRtbUserEvent' as needed to respond to the desired
    published event. 
 ----------------------------------------------------------------------*/

/*
Roundtable Published Events and Parameters:

"addObjectBefore"
  Pcontext = ""
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  Pok      = False will cancel

"addObject"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID

"addObjectToLabBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  Pok      = False with cancel
  
"addObjectToLab"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  
"assignObjectBefore"
  Pother   = Object Name, Object Type, Product Module, Version, Workspace ID
  Pok      = False will cancel

"assignObject"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID

"buildImportListBefore"
  Pcontext = blank
  Pother   = Target Workspace
  Pok      = False will cancel
    
"buildImportList"
  Pcontext = blank
  Pother   = Target Workspace

"buildSchemaListBefore"
  Pcontext = blank
  Pother   = Database Name
  Pok      = False will cancel
    
"buildSchemaList"
  Pcontext = blank
  Pother   = Database Name

"changeObjectShareStatusBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Old Status, New Status, WorkspacePath, Task Path
  Pok      = False will cancel

"changeObjectShareStatus"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Old Status, New Status, WorkspacePath, Task Path

"changePropathBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = "BEFORE-COMPILE" or "AFTER-COMPILE"

"changePropath"
  Pcontext = rtb_object ROWID string value
  Pother   = "BEFORE-COMPILE" or "AFTER-COMPILE"

"changeWorkspaceBefore"
  Pcontext = Desktop handle string value
  Pother   = From-Workspace, To-Workspace
  Pok      = False will cancel

"changeTask"
  Pother   = Task Number

"changeWorkspace"
  Pcontext = Desktop handle string value
  Pother   = Workspace ID

"checkinObjectBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  Pok      = False will cancel

"checkinObject"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID

"checkoutObjectBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID, Level ("version","revision","patch")
  Pok      = False will cancel

"checkoutObject"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID

"compileObjectBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  Pok      = False will cancel

"compileObject"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  
"compileRecordCreate"
  Pcontext = Buffer handle of compile record

"completeTaskBefore"
  Pother   = Task Number
  Pok      = False will cancel

"completeTask"
  Pother   = Task Number

"createObjectVariantBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Current Product Module, New Produce Module, Workspace ID
  Pok      = False will cancel

"createObjectVariant"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID

"createReleaseBefore"
  Pother   = Workspce ID
  Pok      = False will cancel

"createRelease"
  Pother   = Workspace ID, Release Number

"createSiteBefore"
  Pother   = Workspace ID, License Type, Site Code
  Pok      = False will cancel

"createSite"
  Pother   = Workspace ID, License Type, Site Code

"createTaskBefore"
  Pcontext = CHR(1)-delimited list of column/value pairs
  Pother   = Workspace ID
  Pok      = False will cancel

"createTask"
  Pcontext = rtb_task ROWID string value
  Pother   = Task Number

"deleteObjectBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  Pok      = False will cancel

"deleteObject"
  Pcontext = WIP version of the deleted object
  Pother   = Object Name, Object Type, Product Module, Workspace ID

"deleteTaskBefore"
  Pcontext = RowObjUpd buffer handle string value
  Pother   = Task Number
  Pok      = False will cancel

"deleteTask"
  Pcontext = rtb_task ROWID string value
  Pother   = Task Number

"loadPartnerBefore"
  Pcontext = Workspace ID
  Pother   = Workspace Path
  Pok      = False will cancel

"loadPartner"
  Pcontext = rtb_wspace ROWID string value
  Pother   = Object Name

"makeDeploymentBefore"
  Pcontext = rtb_site ROWID string value
  Pother   = rtb_deploy ROWID string value
  Pok      = False will cancel

"makeDeployment"
  Pcontext = rtb_site ROWID string value
  Pother   = rtb_deploy ROWID string value

"modifyTaskBefore"
  Pcontext = RowObjUpd buffer handle string value
  Pother   = Task Number
  Pok      = False will cancel

"modifyTask"
  Pcontext = rtb_task ROWID string value
  Pother   = Task Number

"moveToWebBefore"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name, Object Type, Pmod, Module
  Pok      = False will cancel

"moveToWeb"
  Pcontext = rtb_object ROWID string value
  Pother   = Object Name

"newTaskRow"
  Pcontext = RowObject buffer handle string value
  Pother   = Workspace ID

"processImportBefore"
  Pcontext = Workspace ID
  Pok      = False will cancel

"processImport"
  Pcontext = Workspace ID

"reopenTaskBefore"
  Pcontext = rtb_task ROWID string value
  Pother   = Task Number

"reopenTask"
  Pcontext = rtb_task ROWID string value
  Pother   = Task Number

"roundtableStartup"
  Pcontext = Tabletop procedure handle string value
  Pother   = Tabletop window handle string value

"roundtableShutdown"

"sessionStartup"
  Pcontext = Session ID string value
  Pok      = False will cancel

"sessionShutdown"

"updateSchemaBefore"
  Pother   = Workspace ID
  Pok      = False will cancel

"updateSchema"
  Pother   = Workspace ID

"xrefObjectBefore"
  Pcontext = buffer-handle of compile record string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID
  Pok      = False will cancel

"xrefObject"
  Pcontext = buffer-handle of compile record string value
  Pother   = Object Name, Object Type, Product Module, Workspace ID

*/

SUBSCRIBE TO "evRtbUserEvent":U ANYWHERE.

PROCEDURE evRtbUserEvent :
/*------------------------------------------------------------------------------
  Purpose:     Roundtable Event Handler 
  Parameters:  Pevent
               Pcontext
               Pother
               Pok
  Notes:       Common procedure for all events.
               RETURN a value from this procedure to return an error message
               to the caller.
------------------------------------------------------------------------------*/
  DEFINE INPUT  PARAMETER Pevent   AS CHARACTER.
  DEFINE INPUT  PARAMETER Pcontext AS CHARACTER.
  DEFINE INPUT  PARAMETER Pother   AS CHARACTER.
  DEFINE OUTPUT PARAMETER Pok      AS LOGICAL INITIAL YES.

  DEFINE VARIABLE cError AS CHARACTER NO-UNDO INITIAL "".

  /* Uncomment for debugging */
  
  /*
  MESSAGE
    "Event:   " Pevent SKIP
    "Context: " Pcontext SKIP
    "Other:   " Pother

    VIEW-AS ALERT-BOX
    BUTTONS OK-CANCEL
    TITLE "Roundtable Event"
    UPDATE Pok.
  */  
  
  RETURN cError.

END PROCEDURE.
