/*
 rtb_inst/schupd2.p
 Update the Progress Schema from Roundtable .df files 
*/
DEFINE INPUT  PARAMETER Pfile-name AS CHARACTER NO-UNDO.
DEFINE OUTPUT PARAMETER Perror     AS CHARACTER NO-UNDO.

DEFINE NEW SHARED BUFFER Bfile  FOR DICTDB._file.
DEFINE NEW SHARED BUFFER Bindex FOR DICTDB._index.
DEFINE NEW SHARED BUFFER Bfield FOR DICTDB._field.

DEFINE VARIABLE Mobject        AS CHARACTER NO-UNDO INITIAL "".
DEFINE VARIABLE Mtoken         AS CHARACTER NO-UNDO INITIAL "" EXTENT 10.
DEFINE VARIABLE Mdisp          AS CHARACTER NO-UNDO.
DEFINE VARIABLE Mscope         AS CHARACTER NO-UNDO.
DEFINE VARIABLE Marea-data     AS CHARACTER NO-UNDO.
DEFINE VARIABLE MhFldBuf       AS HANDLE    NO-UNDO.
DEFINE VARIABLE MhSeqBuf       AS HANDLE    NO-UNDO.
DEFINE VARIABLE MhFileBuf      AS HANDLE    NO-UNDO.
DEFINE VARIABLE MhIndexBuf     AS HANDLE    NO-UNDO.
DEFINE VARIABLE Mi             AS INTEGER   NO-UNDO.
DEFINE VARIABLE Mline          AS INTEGER   NO-UNDO.
DEFINE VARIABLE Msequence      AS INTEGER   NO-UNDO.
DEFINE VARIABLE Mfieldcount    AS INTEGER   NO-UNDO.
DEFINE VARIABLE Mk             AS INTEGER   NO-UNDO.
DEFINE VARIABLE Mfo            AS INTEGER   NO-UNDO.
DEFINE VARIABLE Mfile-pos      AS INTEGER   NO-UNDO.
DEFINE VARIABLE Mold-frozen    AS LOGICAL   NO-UNDO.
DEFINE VARIABLE Mfreeze        AS LOGICAL   NO-UNDO.
DEFINE VARIABLE Mfound-token   AS LOGICAL   NO-UNDO.
DEFINE VARIABLE Marea-db       AS LOGICAL   NO-UNDO INITIAL FALSE.
DEFINE VARIABLE Mmulti-db      AS LOGICAL   NO-UNDO INITIAL FALSE.
DEFINE VARIABLE MmultiTenant   AS LOGICAL   NO-UNDO INITIAL FALSE. /* Is table update multi-tenant? */
DEFINE VARIABLE MdefaultArea   AS LOGICAL   NO-UNDO INITIAL FALSE. /* Should default area be used for multi-tenant? */
DEFINE VARIABLE Mpartition-db  AS LOGICAL   NO-UNDO.
DEFINE VARIABLE MisLocal       AS LOGICAL   NO-UNDO.
DEFINE VARIABLE MisPartitioned AS LOGICAL   NO-UNDO.
DEFINE VARIABLE MhasIdxAttrs   AS LOGICAL   NO-UNDO. /* _Index has _index-attributes field - introducted in OE 11.4 */

DEFINE BUFFER B_db       FOR DICTDB._db.
DEFINE BUFFER B2field    FOR DICTDB._field.
DEFINE BUFFER Bifield    FOR DICTDB._index-field.
DEFINE BUFFER Bsequence  FOR DICTDB._sequence.
DEFINE BUFFER Bfiltrig   FOR DICTDB._file-trig.
DEFINE BUFFER Bfldtrig   FOR DICTDB._field-trig.
DEFINE BUFFER Bdbfeature FOR DICTDB._database-feature.

DEFINE SHARED STREAM Supdlog.

DEFINE STREAM Sin1.
DEFINE STREAM Sout1.

DEFINE TEMP-TABLE Tareas
   FIELD dbase      AS CHARACTER
   FIELD table-name AS CHARACTER
   FIELD object     AS CHARACTER
   FIELD obj-type   AS CHARACTER
   FIELD area-name  AS CHARACTER
   FIELD df-file    AS CHARACTER
   INDEX Tindex
         table-name ASCENDING
         object     ASCENDING
         obj-type   ASCENDING.

/*
 Load any applicable area assignments for DB
*/
IF INTEGER(DBVERSION("DICTDB")) >= 9 THEN DO:
  Marea-db = TRUE.
  FILE-INFO:FILE-NAME = "rtb_idat~/dbareas".
  IF FILE-INFO:PATHNAME <> ? THEN DO:
    INPUT STREAM Sin1 FROM VALUE("rtb_idat~/dbareas") NO-ECHO.
    REPEAT:
      CREATE Tareas.
      IMPORT STREAM Sin1 Tareas.
      IF INDEX(Pfile-name,Tareas.df-file) = 0 THEN
        DELETE Tareas.
    END.
    INPUT STREAM Sin1 CLOSE.
  END.
END.

IF INTEGER(DBVERSION("DICTDB")) >= 11 THEN DO:

  /* Check if target is multi-tenant enabled */
  FIND Bdbfeature NO-LOCK
    WHERE Bdbfeature._dbfeature-id = 14 NO-ERROR.
  IF AVAILABLE Bdbfeature THEN
    Mmulti-db = (Bdbfeature._dbfeature_active = "1").

  /* Check if target is partition-enabled */
  FIND Bdbfeature NO-LOCK
    WHERE Bdbfeature._dbfeature-id = 21 NO-ERROR.
  IF AVAILABLE Bdbfeature THEN
    Mpartition-db = (Bdbfeature._dbfeature_active = "1").

END.

INPUT STREAM Sin1 FROM VALUE(Pfile-name) NO-ECHO.

IF NOT SESSION:BATCH-MODE THEN DO:
  PAUSE 0 BEFORE-HIDE.
  HIDE ALL NO-PAUSE.
END.

$TRANS-SCOPE:
DO FOR Bfile, Bindex, Bfield, Bifield WHILE TRUE TRANSACTION ON ERROR UNDO, LEAVE:

  ASSIGN 
    Mscope      = "".     /* Name of _file currently scoped in transaction */
    Mfieldcount = 100000. /* field sequence for field renumbering support */

  $READ-DDL:
  DO WHILE TRUE:
  
    IF TRIM(Mtoken[1]) = "" THEN DO:
      
      Mtoken = "".
      
      DO ON ENDKEY UNDO, LEAVE $TRANS-SCOPE:
        IMPORT STREAM Sin1 Mtoken[1 FOR 10]. 
      END.
      
      Mline = Mline + 1.
      
      Mfound-token = NO.
      
      DO Mi = 1 TO 10:
        Mfound-token = Mfound-token OR TRIM(Mtoken[Mi]) <> "".  /* Mtoken[Mi] = TRIM(Mtoken[Mi]). */
      END.
    
      IF NOT Mfound-token THEN NEXT.   /* skip blank lines */
    
    END. /* read new token */
  
  
    DISPLAY
      Mline FORMAT ">>>>9"
      TRIM(Mtoken[1]) + " " 
      + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])) + " "
      + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])) + " "
      + (IF Mtoken[4] = ? THEN "?" ELSE TRIM(Mtoken[4])) + " "
      + (IF Mtoken[5] = ? THEN "?" ELSE TRIM(Mtoken[5])) + " "
      + (IF Mtoken[6] = ? THEN "?" ELSE TRIM(Mtoken[6])) + " "
      + (IF Mtoken[7] = ? THEN "?" ELSE TRIM(Mtoken[7])) + " "
      + (IF Mtoken[8] = ? THEN "?" ELSE TRIM(Mtoken[8])) + " "
      + (IF Mtoken[9] = ? THEN "?" ELSE TRIM(Mtoken[9])) + " "
      + (IF Mtoken[10] = ? THEN "?" ELSE TRIM(Mtoken[10])) + " "
      @ Mdisp FORMAT "x(72)"
        WITH FRAME  monitor
        NO-ATTR-SPACE
        NO-LABELS
        18 DOWN
        TITLE "---------- Loading Schema Changes ------------".
  
    DOWN 1 WITH FRAME monitor.
  
    /*
     Detect start of new statement
    */
    IF CAN-DO("CREATE,ADD,UPDATE,ALTER,CHANGE,DELETE,DROP,REMOVE,RENAME",TRIM(Mtoken[1])) THEN DO:
  
      IF NOT CAN-DO("SEQUENCE,TABLE,FILE,FIELD,INDEX,UNIQUE,INACTIVE,PRIMARY",TRIM(Mtoken[2])) THEN DO:
        Perror = "Bad object type: " + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])).
        LEAVE $TRANS-SCOPE.
      END. 
  
      IF CAN-DO("UNIQUE,INACTIVE,PRIMARY,INDEX",TRIM(Mtoken[2])) THEN
        Mobject = "INDEX".
      ELSE
        Mobject = TRIM(Mtoken[2]).
  
    END. 
    ELSE DO:
      Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
      LEAVE $TRANS-SCOPE.
    END.
  
  
    /******** S E Q U E N C E    O B J E C T  P R O C E S S I N G ******** 
    Syntax:
    {ADD|CREATE} SEQUENCE "seq-name"
          INITIAL value
          INCREMENT value
          CYCLE-ON-LIMIT {yes|no}
          MIN-VAL value
          MAX-VAL value
  
    {DROP|DELETE} SEQUENCE "seq-name"
  
    {MODIFY|UPDATE|ALTER} SEQUENCE "seq-name"
           INITIAL value
           INCREMENT value
           CYCLE-ON-LIMIT {yes|no}
           MIN-VAL value
           MAX-VAL value
    */
    IF Mobject = "SEQUENCE" THEN DO:
  
      IF Mscope <> "" THEN NEXT $TRANS-SCOPE.
  
      Mtoken[1] = TRIM(Mtoken[1]).
  
      IF CAN-DO("ADD,CREATE",Mtoken[1]) THEN DO:
  
        IF CAN-FIND(FIRST Bsequence 
                      WHERE Bsequence._seq-name = TRIM(Mtoken[3])) THEN DO:
          
          Perror = "Sequence already exists: " + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          LEAVE $TRANS-SCOPE.
        
        END.
  
        FIND FIRST B_db NO-LOCK.
  
        CREATE Bsequence.
        ASSIGN
          Bsequence._db-recid  = RECID(B_db)
          Bsequence._seq-name  = TRIM(Mtoken[3]).
           
        ASSIGN 
          Mtoken = "".
          Mscope = TRIM(Mtoken[3]).
  
      END. /* = "Add" */
  
      ELSE IF CAN-DO("UPDATE,CHANGE,ALTER",Mtoken[1]) THEN DO:
      
        FIND Bsequence 
          WHERE Bsequence._seq-name = TRIM(Mtoken[3]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bsequence THEN DO:
        
          IF LOCKED Bsequence THEN
            Perror = "Sequence record locked: ".
          ELSE
            Perror = "Bad sequence name: ".
         
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        END.
  
        Mscope = TRIM(Mtoken[3]).
  
      END. /* of Update */
  
      ELSE IF CAN-DO("DELETE,DROP,REMOVE",Mtoken[1]) THEN DO:
      
        FIND Bsequence 
          WHERE Bsequence._seq-name = TRIM(Mtoken[3]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bsequence THEN DO:
              
          IF LOCKED Bsequence THEN
            Perror = "Sequence record locked: ".
          ELSE
            Perror = "Bad sequence name: ".
          
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        
        END.
  
        DELETE Bsequence.
  
        Mtoken = "".
        NEXT $TRANS-SCOPE.
      
      END. /* of Delete */
  
      ELSE IF Mtoken[1] = "RENAME" THEN DO:
      
        FIND Bsequence 
          WHERE Bsequence._seq-name = TRIM(Mtoken[3]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bsequence THEN DO:
        
          IF LOCKED Bsequence THEN
            Perror = "Sequence record is locked: ".
          ELSE
            Perror = "Bad sequence name: ".
          
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          
        END.
  
        Bsequence._seq-name = TRIM(Mtoken[5]).
  
        Mtoken = "".
        NEXT $TRANS-SCOPE.
        
      END. /* of Rename */
  
      ELSE DO:
        Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
        UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
      END.
      
      /*
       Process sequence details
      */
      MhSeqBuf = BUFFER Bsequence:HANDLE.
  
      DO WHILE TRUE:
      
        Mtoken = "".
        
        DO ON ENDKEY UNDO, LEAVE $TRANS-SCOPE:
          IMPORT STREAM Sin1 Mtoken[1 FOR 10]. 
        END.
        
        Mline = Mline + 1.
        
        Mfound-token = NO.
        DO Mi = 1 TO 10:
          Mfound-token = Mfound-token OR TRIM(Mtoken[Mi]) <> "". /* Mtoken[Mi] = TRIM(Mtoken[Mi]). */
        END.
           
        IF NOT Mfound-token THEN NEXT. /* skip blank lines */
  
        IF CAN-DO("CREATE,ADD,UPDATE,ALTER,CHANGE,DELETE,DROP,REMOVE,RENAME",TRIM(Mtoken[1])) THEN
          NEXT $READ-DDL .  
  
        /*
         Write token to log
        */  
        DISPLAY
          Mline
          TRIM(Mtoken[1]) + " " 
          + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])) + " "
          + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])) + " "
          + (IF Mtoken[4] = ? THEN "?" ELSE TRIM(Mtoken[4])) + " "
          + (IF Mtoken[5] = ? THEN "?" ELSE TRIM(Mtoken[5])) + " "
          + (IF Mtoken[6] = ? THEN "?" ELSE TRIM(Mtoken[6])) + " "
          + (IF Mtoken[7] = ? THEN "?" ELSE TRIM(Mtoken[7])) + " "
          + (IF Mtoken[8] = ? THEN "?" ELSE TRIM(Mtoken[8])) + " "
          + (IF Mtoken[9] = ? THEN "?" ELSE TRIM(Mtoken[9])) + " "
          + (IF Mtoken[10] = ? THEN "?" ELSE TRIM(Mtoken[10])) + " "
          @ Mdisp
          WITH FRAME monitor.
        
        DOWN 1 WITH FRAME monitor.
        
        /*
         Process token details
        */
        Mtoken[1] = TRIM(Mtoken[1]).
         
        CASE Mtoken[1]:
           
        WHEN "INITIAL" THEN
          Bsequence._seq-init = IF Mtoken[2] = "?" THEN ? ELSE INTEGER(Mtoken[2]). 
         
        WHEN "INCREMENT" THEN
          Bsequence._seq-incr = INTEGER(Mtoken[2]).
         
        WHEN "MIN-VAL" THEN
          Bsequence._seq-min = INTEGER(Mtoken[2]).
         
        WHEN "MAX-VAL" THEN
          Bsequence._seq-max = INTEGER(Mtoken[2]).
         
        WHEN "CYCLE-ON-LIMIT" THEN 
          Bsequence._cycle-ok = (Mtoken[2] = "yes").
         
        WHEN "MULTITENANT" THEN DO:
          IF Mmulti-db THEN
            MhSeqBuf:BUFFER-FIELD("_seq-attributes"):BUFFER-VALUE(1) = YES.
        END.
         
        OTHERWISE DO:
          Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
          LEAVE $TRANS-SCOPE.
        END.
         
        END CASE.
  
        Mtoken = "".
        
      END. /* DO WHILE TRUE  */
    
    END. /* Mobject = "SEQUENCE" */
  
    /******** F I L E  O B J E C T  P R O C E S S I N G ********
    Syntax:
    {ADD|CREATE} {FILE|TABLE} "filename" TYPE PROGRESS 
          DUMP-NAME "dumpname"
          [DESCRIPTION "description"]
          [VALEXP "valexp"]
          [VALMSG "valmessage"]
          [VALMSG-SA ""]
          [FROZEN]
          [HIDDEN]
          [CAN-CREATE ""]
          [CAN-READ ""]
          [CAN-WRITE ""]
          [CAN-DELETE ""]
          [LABEL ""]
          [LABEL-SA ""]
          [FILE-MISC26 "Table Replication program name"]
          [TABLE-TRIGGER "event" DELETE]
          [TABLE-TRIGGER "event" {OVERRIDE|NO-OVERRIDE} PROCEDURE "" CRC "?"] 
   
   {CHANGE|UPDATE|ALTER} {FILE|TABLE} "filename" TYPE PROGRESS
          [DESCRIPTION "description"]
          [VALEXP "valexp"]
          [VALMSG "valmessage"]
          [VALMSG-SA ""]
          [FROZEN]
          [HIDDEN]
          [CAN-CREATE ""]
          [CAN-READ ""]
          [CAN-WRITE ""]
          [CAN-DELETE ""]
          [LABEL ""]
          [LABEL-SA ""]
          [DUMP-NAME "dumpname"]
          [FILE-MISC26 "Table Replication program name"]
          [TABLE-TRIGGER "event" DELETE]
          [TABLE-TRIGGER "event" {OVERRIDE|NO-OVERRIDE} PROCEDURE "" CRC "?"] 
  
    {DELETE|DROP|REMOVE} {FILE|TABLE} "filename" 
  
    RENAME {FILE|TABLE} "filename" TO "newfilename"
    */
    IF CAN-DO("FILE,TABLE",Mobject) THEN DO:
  
      IF Mscope <> "" AND Mscope <> Bfile._file-name THEN
        NEXT $TRANS-SCOPE. 
  
      Mtoken[1] = TRIM(Mtoken[1]).
  
      IF CAN-DO("ADD,CREATE",Mtoken[1]) THEN DO:
  
        IF CAN-FIND(FIRST Bfile
                      WHERE Bfile._file-name = TRIM(Mtoken[3])
                        AND Bfile._DB-lang = 0 AND Bfile._Tbl-type = "T") THEN DO:
              
          Perror = "File already exists: " + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          LEAVE $TRANS-SCOPE.
          
        END.
  
        FIND FIRST B_db NO-LOCK.
  
        CREATE Bfile.
        ASSIGN
          Bfile._db-recid     = RECID(B_db)
          Bfile._file-name    = TRIM(Mtoken[3]).
          
        Mtoken = "".
  
        /*
         Assign to storage area, if applicable
        */
        IF Marea-db THEN DO:
        
          FIND Tareas
            WHERE Tareas.table-name = Bfile._file-name
              AND Tareas.object     = Bfile._file-name
              AND CAN-DO("FILE,TABLE",Tareas.obj-type) NO-ERROR.
             
          IF AVAILABLE Tareas THEN DO:
            
            RUN rtb_inst/schupd2a.p("File",Tareas.area-name).
            
            IF RETURN-VALUE <> "" THEN DO:
              Perror = "Bad Area for File: " + TRIM(Bfile._file-name) + " - Area: " + TRIM(Tareas.area-name).
              LEAVE $TRANS-SCOPE.
            END.
            ELSE
              DELETE Tareas.
             
          END.
        
        END. /* Marea-db */
           
        Mscope = TRIM(Mtoken[3]).
  
      END. /* = "Add" */
  
      ELSE IF CAN-DO("UPDATE,CHANGE,ALTER",Mtoken[1]) THEN DO:
      
        FIND Bfile 
          WHERE Bfile._file-name = TRIM(Mtoken[3])
            AND Bfile._DB-lang   = 0
            AND Bfile._Tbl-type  = "T" EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bfile THEN DO:
        
          IF LOCKED Bfile THEN
            Perror = "File record locked: ".
          ELSE
            Perror = "Bad file name: ".
          
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          
        END.
  
        /*
         Unlock the file so we can update it.  Note that I do not save
         the status of the frozen flag.  Due to a limitation in the 
         Progress DDL we must assume that the FROZEN and HIDDEN flags
         will always be set by any UPDATE to the file definition.  The
         RTB system will do this.  The incremental .df utility in 
         Progress will not.  Note that the incremental would not have
         allowed the udpate anyway, you would have had to manually
         unfreeze the database before loading the .df.
        */
        ASSIGN
          Bfile._hidden = NO.
          Bfile._frozen = NO.
  
        Mscope = TRIM(Mtoken[3]).
  
      END. /* of Update */
  
      ELSE IF CAN-DO("DELETE,DROP,REMOVE",Mtoken[1]) THEN DO:
      
        FIND Bfile 
          WHERE Bfile._file-name = TRIM(Mtoken[3])
            AND Bfile._DB-lang   = 0
            AND Bfile._Tbl-type  = "T" EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bfile THEN DO:
        
          IF LOCKED Bfile THEN
            Perror = "File record locked: ".
          ELSE
            Perror = "Bad file name: ".
          
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        
        END.
  
        Bfile._frozen = NO.
  
        FOR EACH Bindex OF Bfile:
          FOR EACH Bifield OF Bindex:
            DELETE Bifield.
          END.
          DELETE Bindex.
        END.
        
        FOR EACH Bfield OF Bfile:
          FOR EACH Bfldtrig OF Bfield:
            DELETE Bfldtrig.
          END.
          DELETE Bfield.
        END.
        
        FOR EACH Bfiltrig OF Bfile:
          DELETE Bfiltrig.
        END.
  
        DELETE Bfile.
  
        Mtoken = "".
        
        NEXT $TRANS-SCOPE.
        
      END. /* of Delete */
  
      ELSE IF Mtoken[1] = "RENAME" THEN DO:
      
        FIND Bfile 
          WHERE Bfile._file-name = TRIM(Mtoken[3])
            AND Bfile._DB-lang   = 0
            AND Bfile._Tbl-type  = "T" EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bfile THEN DO:
        
          IF LOCKED Bfile THEN
            Perror = "File record is locked: ".
          ELSE
            Perror = "Bad file name: ".
          
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          
        END.
  
        Mold-frozen = Bfile._frozen.
        
        ASSIGN 
          Bfile._frozen    = NO
          Bfile._file-name = TRIM(Mtoken[5])
          Bfile._frozen    = Mold-frozen.
  
        Mtoken = "".
        
        NEXT $TRANS-SCOPE.
        
      END. /* of Rename */
  
      ELSE DO:
        Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
        UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
      END.
  
      /*
       Process table details
      */
      ASSIGN
        Mmultitenant   = NO       /* These should be reset before we process each table */
        MdefaultArea   = NO
        MisPartitioned = NO.
        
      MhFileBuf = BUFFER Bfile:HANDLE.
  
      TABLE-UPDATE:
      DO WHILE TRUE:
      
        Mtoken = "".
        DO ON ENDKEY UNDO, LEAVE TABLE-UPDATE:
          IMPORT STREAM Sin1 Mtoken[1 FOR 10]. 
        END.
        
        Mline = Mline + 1.
        
        Mfound-token = NO.
        DO Mi = 1 TO 10:
          Mfound-token = Mfound-token OR TRIM(Mtoken[Mi]) <> "". /* Mtoken[Mi] = TRIM(Mtoken[Mi]). */
        END.
         
        IF NOT Mfound-token THEN NEXT.   /* skip blank lines */
        
        IF CAN-DO("CREATE,ADD,UPDATE,ALTER,CHANGE,DELETE,DROP,REMOVE,RENAME",TRIM(Mtoken[1])) THEN DO:
          
          /*
           Done with TABLE - set multi-tenant info before moving on..
          */
          IF MmultiTenant THEN DO:
            MhFileBuf:BUFFER-FIELD("_file-attributes"):BUFFER-VALUE(1) = YES.
            IF MdefaultArea THEN
              MhFileBuf:BUFFER-FIELD("_file-attributes"):BUFFER-VALUE(2) = YES.
          END.
          
          /* set table partitioned before moving on... */
          IF MisPartitioned THEN DO:
            
            IF Mpartition-db THEN
              MhFileBuf:BUFFER-FIELD("_file-attributes"):BUFFER-VALUE(3) = YES.
            ELSE
              PUT STREAM Supdlog UNFORMATTED
                 SUBSTITUTE("**Table '&1' defined as partitioned, but DB is not partition-enabled. Property not set.", MhFileBuf::_File-name) + CHR(10).

          END.

          NEXT $READ-DDL.
  
        END.
        
        /*
         Write token to log
        */
        DISPLAY
          Mline
          TRIM(Mtoken[1]) + " " 
          + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])) + " "
          + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])) + " "
          + (IF Mtoken[4] = ? THEN "?" ELSE TRIM(Mtoken[4])) + " "
          + (IF Mtoken[5] = ? THEN "?" ELSE TRIM(Mtoken[5])) + " "
          + (IF Mtoken[6] = ? THEN "?" ELSE TRIM(Mtoken[6])) + " "
          + (IF Mtoken[7] = ? THEN "?" ELSE TRIM(Mtoken[7])) + " "
          + (IF Mtoken[8] = ? THEN "?" ELSE TRIM(Mtoken[8])) + " "
          + (IF Mtoken[9] = ? THEN "?" ELSE TRIM(Mtoken[9])) + " "
          + (IF Mtoken[10] = ? THEN "?" ELSE TRIM(Mtoken[10])) + " "
          @ Mdisp
          WITH FRAME monitor.
        
        DOWN 1 WITH FRAME monitor.
        
        /*
         Process token details
        */
        Mtoken[1] = TRIM(Mtoken[1]).
        
        CASE Mtoken[1]:
  
        WHEN "DESCRIPTION" THEN
          Bfile._desc = Mtoken[2]. 
        
        WHEN "VALEXP" THEN
          Bfile._valexp = Mtoken[2].
         
        WHEN "VALMSG" THEN
          Bfile._valmsg = Mtoken[2].
         
        WHEN "FROZEN" THEN
          Mfreeze = YES.
        
        WHEN "HIDDEN" THEN
          Bfile._hidden = YES.
        
        WHEN "CAN-CREATE" THEN
          Bfile._can-create = Mtoken[2].
        
        WHEN "CAN-READ" THEN
          Bfile._can-read = Mtoken[2].
        
        WHEN "CAN-WRITE" THEN
          Bfile._can-write = Mtoken[2].
        
        WHEN "CAN-DELETE" THEN
          Bfile._can-delete = Mtoken[2].
        
        WHEN "DUMP-NAME" THEN
          Bfile._dump-name = TRIM(Mtoken[2]).
        
        WHEN "FILE-MISC26" THEN
          Bfile._fil-misc2[6] = TRIM(Mtoken[2]).
        
        WHEN "VALMSG-SA" THEN
          Bfile._valmsg-sa = Mtoken[2].
        
        WHEN "LABEL" THEN
          Bfile._file-label = Mtoken[2].
        
        WHEN "LABEL-SA" THEN
          Bfile._file-label-sa = Mtoken[2].
        
        WHEN "AREA" THEN DO:
        END.  /* Ignore AREA token. Already handled. */
           
        WHEN "MULTITENANT" THEN DO:
          IF Mmulti-db THEN
            ASSIGN
              MmultiTenant = YES
              MdefaultArea = YES.
        END.
        
        WHEN "IS-PARTITIONED" THEN DO:
          IF Mtoken[2] = "TRUE" THEN
             MisPartitioned = YES.
        END.
           
        WHEN "NO-DEFAULT-AREA" THEN DO:
          IF Mmulti-db THEN
            MdefaultArea = NO.
        END.
           
        WHEN "TABLE-TRIGGER" THEN DO:
        
          FIND Bfiltrig OF Bfile 
            WHERE Bfiltrig._event = Mtoken[2] EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
          
          /*
           Progress doesn't let you modify a trigger record, so delete and
           recreate, as required.
          */
          IF AVAILABLE Bfiltrig THEN
            DELETE Bfiltrig.
            
          IF Mtoken[3] <> "DELETE" THEN DO:
            
            CREATE Bfiltrig.
            ASSIGN
              Bfiltrig._file-recid = RECID(Bfile)
              Bfiltrig._event = Mtoken[2].
  
            Mk = 3.
            DO WHILE Mk < 5:
            
              IF Mtoken[Mk] = "NO-OVERRIDE" THEN
                Bfiltrig._override = NO.
              
              ELSE IF Mtoken[Mk] = "OVERRIDE" THEN
                Bfiltrig._override = YES.
              
              ELSE IF Mtoken[Mk] = "PROCEDURE" THEN 
                ASSIGN
                  Mk = Mk + 1
                  Bfiltrig._proc-name = Mtoken[Mk].
              
              ELSE IF Mtoken[Mk] = "CRC" THEN 
                ASSIGN
                  Mk = Mk + 1
                  Bfiltrig._trig-crc = INTEGER(Mtoken[Mk]).
                 
              Mk = Mk + 1.
              
            END.
            
          END.
          
        END. /* TABLE-TRIGGER */
           
        OTHERWISE DO:
          Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
          LEAVE $TRANS-SCOPE.
        END.
        
        END CASE.
  
        Mtoken = "".
  
      END. /* TABLE-UPDATE  */
  
      /*
       Set multi-tenant properties.
      */ 
      IF MmultiTenant THEN DO:
        MhFileBuf:BUFFER-FIELD("_file-attributes"):BUFFER-VALUE(1) = YES.
        IF MdefaultArea THEN
          MhFileBuf:BUFFER-FIELD("_file-attributes"):BUFFER-VALUE(2) = YES.
      END.
        
      /*
       Set partitioned property
      */
      IF MisPartitioned THEN DO:
        IF Mpartition-db THEN
          MhFileBuf:BUFFER-FIELD("_file-attributes"):BUFFER-VALUE(3) = YES.
        ELSE
          PUT STREAM Supdlog UNFORMATTED
            SUBSTITUTE("**Table '&1' defined as partitioned, but DB is not partition-enabled. Table placed in Schema Area.", MhFileBuf::_File-name) + CHR(10).
      END.

      IF Mfreeze THEN 
        ASSIGN
          Bfile._frozen = YES 
          Mfreeze       = NO.
        
      LEAVE $TRANS-SCOPE.
  
    END. /* Mobject = "FILE" */
  
    /********  F I E L D   O B J E C T  P R O C E S S I N G ********
     Syntax:
     {ADD|CREATE} FIELD "fieldname" OF "filename" TYPE {character|integer|logical|decimal|recid|date}
          FORMAT ""
          EXTENT 99
          ORDER  9999
          [FORMAT-SA]
          [DESCRIPTION ""]
          [HELP ""]
          [HELP-SA ""]
          [INITIAL ""]
          [INITIAL-SA ""]
          [LABEL ""]
          [LABEL-SA ""]
          [COLUMN-LABEL ""]
          [COLUMN-LABEL-SA ""]
          [CAN-READ ""]
          [CAN-WRITE ""]
          [VALEXP ""]
          [VALMSG ""]
          [VALMSG-SA ""]
          [DECIMALS 99]
          [{MANDATORY|NULL-ALLOWED}]
          [{CASE-SENSITIVE|NOT-CASE-SENSITIVE}]
          [FIELD-TRIGGER "event" {OVERRIDE|NO-OVERRIDE} PROCEDURE "" CRC "?"] 
          
     {CHANGE|UPDATE|ALTER} FIELD "fieldname" OF "filename"
         [FORMAT]
         [FORMAT-SA]
         [ORDER]
         [DESCRIPTION ""]
         [HELP ""]
         [HELP-SA ""]
         [INITIAL ""]
         [INITIAL-SA ""]
         [LABEL ""]
         [LABEL-SA ""]
         [COLUMN-LABEL ""]
         [COLUMN-LABEL-SA ""]
         [CAN-READ ""]
         [CAN-WRITE ""]
         [VALEXP ""]
         [VALMSG ""]
         [VALMSG-SA ""]
         [DECIMALS 99]
         [{MANDATORY|NULL-ALLOWED}]
         [{CASE-SENSITIVE|NOT-CASE-SENSITIVE}]
         [FIELD-TRIGGER "event" DELETE]
         [FIELD-TRIGGER "event" {OVERRIDE|NO-OVERRIDE} PROCEDURE "" CRC "?"] 
  
     {DELETE|DROP|REMOVE} FIELD "fieldname" OF "filename"
  
     RENAME FIELD "fieldname" OF "fielname" TO "newfieldname"
    */
    IF Mobject = "FIELD" THEN DO:
  
      FIND Bfile
        WHERE Bfile._file-name = TRIM(Mtoken[5])
          AND Bfile._DB-lang   = 0
          AND Bfile._Tbl-type  = "T" EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
      IF NOT AVAILABLE Bfile THEN DO:
      
        IF LOCKED Bfile THEN
          Perror = "File record locked: ".
        ELSE
          Perror = "Bad file name: ".
          
        Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
        LEAVE $TRANS-SCOPE.
        
      END.
  
      /*
       Unlock file so we can update it
      */
      ASSIGN 
        Mold-frozen   = Bfile._frozen.
        Bfile._frozen = NO.
  
      IF Mscope <> "" AND Mscope <> Bfile._file-name THEN
        NEXT $TRANS-SCOPE. 
  
      IF Mscope = "" THEN  /* lock all fields of file */
        FOR EACH Bfield OF Bfile EXCLUSIVE-LOCK:
        END.
  
      Mscope = Bfile._file-name.
  
      IF CAN-DO("ADD,CREATE",Mtoken[1]) THEN DO:
  
        FIND Bfield OF Bfile 
          WHERE Bfield._field-name = TRIM(Mtoken[3]) NO-LOCK NO-ERROR.
  
        IF AVAILABLE Bfield THEN DO:
          Perror = "Field already exists: " + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        END.
  
        CREATE Bfield.
      
        ASSIGN
          Bfield._field-name  = TRIM(Mtoken[3])
          Bfield._file-recid  = RECID(Bfile)
          Bfield._data-type   = TRIM(Mtoken[7]).
  
        /*
         Assign to storage area, if applicable
        */
        IF Marea-db THEN DO:
      
          FIND Tareas
            WHERE Tareas.table-name = Bfile._file-name
              AND Tareas.object     = Bfield._Field-name
              AND Tareas.obj-type   = "LOB" NO-ERROR.
        
          IF AVAILABLE Tareas THEN DO:
          
            RUN rtb_inst/schupd2a.p("LOB",Tareas.area-name).
          
            IF RETURN-VALUE <> "" THEN DO:
              Perror = "Bad Area for LOB: " + TRIM(Bfield._Field-name) + " - Area: " + TRIM(Tareas.area-name).
              LEAVE $TRANS-SCOPE.
            END.
            ELSE
              DELETE Tareas.
             
          END.
        
        END.
      
      END. /* "ADD,CREATE" */
  
      ELSE IF CAN-DO("UPDATE,ALTER,CHANGE",Mtoken[1]) THEN DO:
  
        FIND Bfield OF Bfile
          WHERE Bfield._field-name = TRIM(Mtoken[3]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bfield THEN DO:
      
          IF LOCKED Bfield THEN
            Perror = "Field record is locked: ".
          ELSE
            Perror = "Bad Field name: ".
          
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        
        END.
  
      END. /* Update */
  
      ELSE IF Mtoken[1] = "RENAME" THEN DO:
  
        FIND Bfield OF Bfile
          WHERE Bfield._field-name = TRIM(Mtoken[3]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bfield THEN DO:
      
          IF LOCKED Bfield THEN
            Perror = "Field record is locked: ".
          ELSE
            Perror = "Bad Field name: ".
        
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        
        END.
  
        Bfield._field-name = TRIM(Mtoken[7]). 
        Mtoken = "".
      
      END. /* Update */
  
      ELSE IF CAN-DO("DELETE,DROP,REMOVE",Mtoken[1]) THEN DO:
    
        FIND Bfield OF Bfile 
          WHERE Bfield._field-name = TRIM(Mtoken[3]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bfield THEN DO:
      
          IF LOCKED Bfield THEN
            Perror = "Field record is locked: ".
          ELSE
            Perror = "Bad Field name: ".
        
          Perror = Perror + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
      
        END.
  
        FOR EACH Bfldtrig OF Bfield:
          DELETE Bfldtrig.
        END.
  
        DELETE Bfield.
  
        Mtoken = "".
      
        NEXT $READ-DDL.
  
      END. /* Delete */
  
      ELSE DO:
        Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
        UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
      END.
  
      /*
       Process field details
      */
      MhFldBuf = BUFFER Bfield:HANDLE.
  
      FIELD-UPDATE:
      DO WHILE TRUE:
      
        Mtoken = "".
        DO ON ENDKEY UNDO, LEAVE FIELD-UPDATE:
          IMPORT STREAM Sin1 Mtoken[1 FOR 10]. 
        END.
       
        Mline = Mline + 1.
       
        Mfound-token = NO.
        DO Mi = 1 TO 10:
          Mfound-token = Mfound-token OR TRIM(Mtoken[Mi]) <> "". /* Mtoken[Mi] = TRIM(Mtoken[Mi]). */
        END.
           
        IF NOT Mfound-token THEN NEXT. /* skip blank lines */
  
        IF CAN-DO("CREATE,ADD,UPDATE,ALTER,CHANGE,DELETE,DROP,REMOVE,RENAME",TRIM(Mtoken[1])) THEN
          NEXT $READ-DDL.
          
        /*
         Log current token
        */  
        DISPLAY
          Mline
          TRIM(Mtoken[1]) + " " 
           + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])) + " "
           + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])) + " "
           + (IF Mtoken[4] = ? THEN "?" ELSE TRIM(Mtoken[4])) + " "
           + (IF Mtoken[5] = ? THEN "?" ELSE TRIM(Mtoken[5])) + " "
           + (IF Mtoken[6] = ? THEN "?" ELSE TRIM(Mtoken[6])) + " "
           + (IF Mtoken[7] = ? THEN "?" ELSE TRIM(Mtoken[7])) + " "
           + (IF Mtoken[8] = ? THEN "?" ELSE TRIM(Mtoken[8])) + " "
           + (IF Mtoken[9] = ? THEN "?" ELSE TRIM(Mtoken[9])) + " "
           + (IF Mtoken[10] = ? THEN "?" ELSE TRIM(Mtoken[10])) + " "
           @ Mdisp
           WITH FRAME monitor.
  
        DOWN 1 WITH FRAME monitor.
      
        /*
         Process token details
        */
        Mtoken[1] = TRIM(Mtoken[1]).
        
        CASE Mtoken[1]:
  
        WHEN "DESCRIPTION" THEN
          Bfield._desc = Mtoken[2].
       
        WHEN "FORMAT" THEN
          Bfield._format = Mtoken[2].
        
        WHEN "SQL-WIDTH" THEN
          Bfield._width = INTEGER(Mtoken[2]).
          
        WHEN "HELP" THEN
          Bfield._help = Mtoken[2].
        
        WHEN "INITIAL" THEN
          Bfield._initial = Mtoken[2].
          
        WHEN "LABEL" THEN
          Bfield._label = Mtoken[2].
        
        WHEN "COLUMN-LABEL" THEN
          Bfield._col-label = Mtoken[2].
        
        WHEN "CAN-READ" THEN
          Bfield._can-read = Mtoken[2].
       
        WHEN "CAN-WRITE" THEN
          Bfield._can-write = Mtoken[2].
        
        WHEN "VALEXP" THEN
          Bfield._valexp = Mtoken[2].
        
        WHEN "VALMSG" THEN
          Bfield._valmsg = Mtoken[2].
        
        WHEN "EXTENT" THEN DO:
          IF NEW Bfield THEN
            Bfield._extent = INTEGER(Mtoken[2]).
        END.
        
        WHEN "DECIMALS" THEN 
          Bfield._decimals = INTEGER(Mtoken[2]).
        
        WHEN "MANDATORY" THEN
          Bfield._mandatory = YES.
        
        WHEN "NULL-ALLOWED" THEN
          Bfield._mandatory = NO.
        
        WHEN "CASE-SENSITIVE" THEN
          Bfield._fld-case = YES.
        
        WHEN "NOT-CASE-SENSITIVE" THEN
          Bfield._fld-case = NO.
        
        WHEN "ORDER" THEN DO:
        
          FIND B2field OF Bfile 
            WHERE B2field._order = INTEGER(Mtoken[2]) EXCLUSIVE-LOCK NO-ERROR.
  
          IF AVAILABLE B2field THEN
            ASSIGN 
              B2field._order = Mfieldcount
              Mfieldcount = Mfieldcount + 1.
  
          Bfield._order = INTEGER(Mtoken[2]).
            
        END. /* ORDER */
        
        WHEN "FORMAT-SA" THEN
          Bfield._format-sa = Mtoken[2].
        
        WHEN "HELP-SA" THEN
          Bfield._help-sa = Mtoken[2].
        
        WHEN "INITIAL-SA" THEN
          Bfield._initial-sa = Mtoken[2].
        
        WHEN "COLUMN-LABEL-SA" THEN
          Bfield._col-label-sa = Mtoken[2].
        
        WHEN "VALMSG-SA" THEN
          Bfield._valmsg-sa = Mtoken[2].
        
        WHEN "VIEW-AS" THEN
          Bfield._view-as = Mtoken[2].
         
        WHEN "LABEL-SA" THEN
          Bfield._label-sa = Mtoken[2].
        
        WHEN "AREA" THEN DO:
        END. /* Ignore AREA token. Already handled. */
        
        WHEN "LOB-SIZE" THEN
          MhFldBuf:BUFFER-FIELD("_Fld-misc2"):BUFFER-VALUE(1) = Mtoken[2].
        
        WHEN "CLOB-CODEPAGE" THEN DO:
          /*
           Use DB codepage if value is unknown
          */
          IF Mtoken[2] = ? THEN DO:
            FIND FIRST B_db WHERE B_db._Db-name = ? NO-LOCK.
            MhFldBuf:BUFFER-FIELD("_Charset":U):BUFFER-VALUE = B_db._Db-xl-name.
          END.
          ELSE
            MhFldBuf:BUFFER-FIELD("_Charset"):BUFFER-VALUE = Mtoken[2] .
        END.
        
        WHEN "CLOB-COLLATION" THEN DO:
          /*
           Use DB collation if value is unknown
          */
          IF Mtoken[2] = ? THEN DO:
            FIND FIRST B_db WHERE B_db._Db-name = ? NO-LOCK.
            MhFldBuf:BUFFER-FIELD("_Collation":U):BUFFER-VALUE = B_db._Db-coll-name.
          END.
          ELSE
            MhFldBuf:BUFFER-FIELD("_Collation"):BUFFER-VALUE = Mtoken[2].
        END.
        
        WHEN "CLOB-TYPE" THEN
          MhFldBuf:BUFFER-FIELD("_Attributes1"):BUFFER-VALUE = INTEGER(Mtoken[2]).
        
        WHEN "FIELD-TRIGGER" THEN DO:
        
          FIND Bfldtrig OF Bfield 
            WHERE Bfldtrig._event = Mtoken[2] EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
          
          IF AVAILABLE Bfldtrig THEN
            DELETE Bfldtrig.
            
          IF Mtoken[3] <> "DELETE" THEN DO:
            
            CREATE Bfldtrig.
            ASSIGN
              Bfldtrig._field-recid = RECID(Bfield)
              Bfldtrig._file-recid  = RECID(Bfile)
              Bfldtrig._event       = Mtoken[2].
  
            Mk = 3.
            DO WHILE Mk < 5:
            
              IF Mtoken[Mk] = "NO-OVERRIDE" THEN
                Bfldtrig._override = NO.
              
              ELSE IF Mtoken[Mk] = "OVERRIDE" THEN
                Bfldtrig._override = YES.
              
              ELSE IF Mtoken[Mk] = "PROCEDURE" THEN 
                ASSIGN
                  Mk = Mk + 1
                  Bfldtrig._proc-name = Mtoken[Mk].
                  
              ELSE IF Mtoken[Mk] = "CRC" THEN 
                ASSIGN
                  Mk = Mk + 1
                  Bfldtrig._trig-crc = INTEGER(Mtoken[Mk]).
                  
              Mk = Mk + 1.
              
            END.
          
          END.
              
        END.
        
        OTHERWISE DO:
          Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
          LEAVE $TRANS-SCOPE.
        END.
        
        END CASE.
  
        Mtoken = "".
  
      END. /* of FIELD-UPDATE */
  
      Bfile._frozen = Mold-frozen.
      
      LEAVE $TRANS-SCOPE.
  
    END. /* Mobject = "FIELD" */
  
    /*******  I N D E X   P R O C E S S I N G *******
     v6 Syntax:
     {ADD|CREATE} [UNIQUE] [INACTIVE] [PRIMARY] INDEX "indexname" ON "filename"
           [INDEX-NUM 99999] 
           INDEX-FIELD "fieldname" {ASCENDING|DESCENDING} [UNSORTED] [ABBREVIATED]
    
     {DELETE|DROP|REMOVE} INDEX "indexname" ON "filename"
    
     {UPDATE|CHANGE|ALTER} [PRIMARY] [INACTIVE] INDEX "indexname" ON "filename"
    
     RENAME INDEX "indexname" TO "newindexname"  ON "filename"
     or for backward compatibility with early versions of roundtable
     RENAME INDEX "indexname" ON "filename" TO "newindexname"
    
     v7 Syntax:
     {ADD|CREATE}INDEX "indexname" ON "filename" 
           [DESCRIPTION] ""
           [WORD]
           [UNIQUE] 
           [INACTIVE] 
           [PRIMARY]
           [INDEX-NUM 99999] 
           INDEX-FIELD "fieldname" {ASCENDING|DESCENDING} [UNSORTED] [ABBREVIATED]
    
     {DELETE|DROP|REMOVE} INDEX "indexname" ON "filename"
    
     {UPDATE|CHANGE|ALTER} INDEX "indexname" ON "filename"
           [DESCRIPTION] ""
           [INACTIVE]
           [PRIMARY]
    
     RENAME INDEX "indexname" TO "newindexname"  ON "filename"
     or for backward compatibility with early versions of roundtable
     RENAME INDEX "indexname" ON "filename" TO "newindexname"
    */
    IF Mobject = "INDEX" THEN DO:
  
      /*
       Find token containing INDEX keyword so we know where to
       grab index name and file name
      */
      DO Mi = 2 TO 5 WHILE TRIM(Mtoken[Mi]) <> "INDEX":
      END.
  
      Mfile-pos = IF TRIM(Mtoken[Mi + 2]) = "ON" THEN
                    Mi + 3
                  ELSE
                    Mi + 5.
  
      FIND Bfile
        WHERE Bfile._file-name = TRIM(Mtoken[Mfile-pos])
          AND Bfile._DB-lang   = 0
          AND Bfile._Tbl-type  = "T" EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
      IF NOT AVAILABLE Bfile THEN DO:
      
        IF LOCKED Bfile THEN
          Perror = "File record is locked: ".
        ELSE
          Perror = "Bad file name: ".
        
        Perror = Perror + (IF Mtoken[Mfile-pos] = ? THEN "?" ELSE TRIM(Mtoken[Mfile-pos])).
        UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
        
      END.
  
      /* Unfreeze the file so we can update it */
      Mold-frozen = Bfile._frozen.
      Bfile._frozen = NO.
  
      IF Mscope = "" THEN
        Mscope = TRIM(Mtoken[Mfile-pos]).
  
      /* If we have changed files then start a new transaction */
      IF Mscope <> Bfile._file-name THEN
        NEXT $TRANS-SCOPE.
  
      Mscope = Bfile._file-name.
  
      IF CAN-DO("ADD,CREATE",Mtoken[1]) THEN DO:
  
        IF TRIM(Mtoken[Mi + 1]) <> "default" THEN DO:
  
          FIND Bindex OF Bfile 
            WHERE Bindex._index-name = TRIM(Mtoken[Mi + 1]) NO-LOCK NO-ERROR.
  
          IF AVAILABLE Bindex THEN DO:
            Perror = "Index already exists: " + (IF Mtoken[Mi + 1] = ? THEN "?" ELSE TRIM(Mtoken[Mi + 1])).
            UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          END.
  
          /*
           Delete "default" index if it exists
          */
          FIND Bindex 
            WHERE Bindex._file-recid = RECID(Bfile) 
              AND Bindex._index-name = "default" EXCLUSIVE-LOCK NO-ERROR.
          
          IF AVAILABLE Bindex THEN DO:
            
            FIND FIRST Bifield
              WHERE Bifield._index-recid = RECID(Bindex) EXCLUSIVE-LOCK NO-ERROR.
              
            IF AVAILABLE Bifield THEN DELETE Bifield.
              DELETE Bindex.
            
          END.            
  
          CREATE Bindex.
          ASSIGN
            Bindex._file-recid   = RECID(Bfile)
            Bindex._index-name   = TRIM(Mtoken[Mi + 1])
            Bindex._unique       = NO 
            Bindex._active       = YES.
  
          DO Mi = 2 TO 5: 
              
            IF TRIM(Mtoken[Mi]) = "UNIQUE" THEN
              Bindex._unique = YES.
              
            IF TRIM(Mtoken[Mi]) = "INACTIVE" THEN
              Bindex._active = NO.
              
            IF TRIM(Mtoken[Mi]) = "PRIMARY" THEN
              Bfile._prime-index = RECID(Bindex).
              
          END.
            
        END.  /* if name <> "default" */
              
        /*
         Assign to storage area, if applicable
        */
        IF Marea-db THEN DO:
          
          FIND Tareas
            WHERE Tareas.table-name = Bfile._file-name
              AND Tareas.object     = Bindex._index-name
              AND Tareas.obj-type   = "INDEX" NO-ERROR.
                
          IF AVAILABLE Tareas THEN DO:
            
            RUN rtb_inst/schupd2a.p("Index",Tareas.area-name).
              
            IF RETURN-VALUE <> "" THEN DO:
              Perror = "Bad Area for Index: " + TRIM(Bindex._index-name) + " - Area: " + TRIM(Tareas.area-name).
              LEAVE $TRANS-SCOPE.
            END.
            ELSE
              DELETE Tareas.
                
          END.
            
        END.
        
      END. /* "ADD,CREATE" */
  
      ELSE IF CAN-DO("UPDATE,CHANGE,ALTER",Mtoken[1]) THEN DO:
      
        FIND Bindex OF Bfile 
          WHERE Bindex._index-name = TRIM(Mtoken[Mi + 1]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bindex THEN DO:
          
          IF LOCKED Bindex THEN
            Perror = "Index record locked: ".
          ELSE
            Perror = "Bad Index name: ".
          
          Perror = Perror + (IF Mtoken[Mi + 1] = ? THEN "?" ELSE TRIM(Mtoken[Mi + 1])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          
        END.
  
        DO Mi = 2 TO 3: 
          
          IF TRIM(Mtoken[Mi]) = "INACTIVE" THEN
            Bindex._active = NO.
          
          IF TRIM(Mtoken[Mi]) = "PRIMARY" THEN
            Bfile._prime-index = RECID(Bindex).
            
        END.
  
        IF TRUE THEN DO:
        END.
        ELSE DO:
          Mtoken = "".
          IF Bfile._prime-index <> -999999 THEN
            NEXT $TRANS-SCOPE.
        END.
  
      END. /* Update */
      
      ELSE IF CAN-DO("RENAME",Mtoken[1]) THEN DO:
  
        FIND Bindex OF Bfile 
          WHERE Bindex._index-name = TRIM(Mtoken[Mi + 1]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bindex THEN DO:
        
          IF LOCKED Bindex THEN
            Perror = "Index record is locked: ".
          ELSE
            Perror = "Bad Index name: ". 
          
          Perror = Perror + (IF Mtoken[Mi + 1] = ? THEN "?" ELSE TRIM(Mtoken[Mi + 1])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          
        END.
  
        IF Mtoken[4] = "ON" THEN
          Bindex._index-name = TRIM(Mtoken[7]).
        ELSE
          Bindex._index-name = TRIM(Mtoken[5]).
  
        Mtoken = "".
        NEXT $READ-DDL.
          
      END. /* Rename */
  
      ELSE IF CAN-DO("DELETE,DROP,REMOVE",Mtoken[1]) THEN DO:
      
        FIND Bindex OF Bfile 
          WHERE Bindex._index-name = TRIM(Mtoken[Mi + 1]) EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
  
        IF NOT AVAILABLE Bindex THEN DO:
        
          IF LOCKED Bindex THEN
            Perror = "Index record is locked: ".
          ELSE
            Perror = "Bad Index name: ".
          
          Perror = Perror + (IF Mtoken[Mi + 1] = ? THEN "?" ELSE TRIM(Mtoken[Mi + 1])).
          UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          
        END.
  
        IF Bfile._prime-index = RECID(Bindex) THEN
          Bfile._prime-index = -999999.
  
        FOR EACH Bifield OF Bindex:
          DELETE Bifield.
        END.
        
        DELETE Bindex.
         
        Mtoken = "".
   
        IF Bfile._prime-index <> -999999 THEN
          NEXT $TRANS-SCOPE.
                  
      END. /* of Delete */
  
      ELSE DO:
        Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
        LEAVE $TRANS-SCOPE.
      END.
  
      /*
       Process index details
      */
      Msequence = 0.
   
      INDEX-UPDATE:
      DO WHILE TRUE:
      
        Mtoken = "".
        DO ON ENDKEY UNDO, LEAVE INDEX-UPDATE:
          IMPORT STREAM Sin1 Mtoken[1 FOR 10]. 
        END.
       
        Mline = Mline + 1.
        
        Mfound-token = NO.
        DO Mi = 1 TO 10:
          Mfound-token = Mfound-token OR TRIM(Mtoken[Mi]) <> "".       /* Mtoken[Mi] = TRIM(Mtoken[Mi]). */
        END.
         
        IF NOT Mfound-token THEN NEXT. /* skip blank lines */
  
        IF CAN-DO("CREATE,ADD,UPDATE,ALTER,CHANGE,DELETE,DROP,REMOVE,RENAME",TRIM(Mtoken[1])) THEN
          NEXT $READ-DDL.
        
        /*
         Log current token
        */  
        DISPLAY
          Mline
          "Isec " +
          TRIM(Mtoken[1]) + " " 
          + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])) + " "
          + (IF Mtoken[3] = ? THEN "?" ELSE TRIM(Mtoken[3])) + " "
          + (IF Mtoken[4] = ? THEN "?" ELSE TRIM(Mtoken[4])) + " "
          + (IF Mtoken[5] = ? THEN "?" ELSE TRIM(Mtoken[5])) + " "
          + (IF Mtoken[6] = ? THEN "?" ELSE TRIM(Mtoken[6])) + " "
          + (IF Mtoken[7] = ? THEN "?" ELSE TRIM(Mtoken[7])) + " "
          + (IF Mtoken[8] = ? THEN "?" ELSE TRIM(Mtoken[8])) + " "
          + (IF Mtoken[9] = ? THEN "?" ELSE TRIM(Mtoken[9])) + " "
          + (IF Mtoken[10] = ? THEN "?" ELSE TRIM(Mtoken[10])) + " "
          @ Mdisp
          WITH FRAME monitor.
  
        DOWN 1 WITH FRAME monitor.
        
        /*
         Process token details
        */
        Mtoken[1] = TRIM(Mtoken[1]).
        
        CASE Mtoken[1]:
          
        WHEN "INDEX-NUM" THEN
          Bindex._idx-num = INTEGER(Mtoken[1]).
          
        WHEN "UNIQUE" THEN
          Bindex._unique = YES.
          
        WHEN "INACTIVE" THEN
          Bindex._active = NO.
          
        WHEN "PRIMARY" THEN
          Bfile._prime-index = RECID(Bindex).
          
        WHEN "WORD" THEN
          Bindex._wordidx = 1.
          
        WHEN "DESCRIPTION" THEN
          Bindex._desc = Mtoken[2].
          
        WHEN "AREA" THEN DO:
        END. /* Ignore AREA token. Already handled */
        
        WHEN "IS-LOCAL" THEN DO:
          
          ASSIGN 
            MhIndexBuf   = BUFFER Bindex:HANDLE
            MhasIdxAttrs = VALID-HANDLE(MhIndexBuf:BUFFER-FIELD("_index-attributes"))
            MisLocal     = LOGICAL(Mtoken[2]).
          
          IF MisPartitioned AND MhasIdxAttrs AND MisLocal THEN DO:
            
            IF Mpartition-db THEN
              MhIndexBuf::_index-attributes(1) = MisLocal.
            ELSE
              PUT STREAM Supdlog UNFORMATTED
                SUBSTITUTE("**Index '&1' defined as Local, but DB is not partition-enabled. Index added as Global to Schema Area.",MhIndexBuf::_index-name) + CHR(10).
          
          END.

        END.
        
        WHEN "INDEX-FIELD" THEN DO:
        
          FIND Bfield OF Bfile 
            WHERE Bfield._field-name   = TRIM(Mtoken[2]) NO-LOCK NO-ERROR.
  
          IF NOT AVAILABLE Bfield THEN DO:
            Perror = "Bad index field name: " + (IF Mtoken[2] = ? THEN "?" ELSE TRIM(Mtoken[2])).
            UNDO $TRANS-SCOPE, LEAVE $TRANS-SCOPE.
          END.
  
          Msequence = Msequence + 1.
          
          CREATE Bifield.
          ASSIGN 
            Bifield._index-recid = RECID(Bindex)
            Bifield._field-recid = RECID(Bfield)
            Bifield._index-seq   = Msequence
            Bifield._ascending   = NO 
            Bifield._abbreviate  = NO 
            Bifield._unsorted    = NO.
  
          DO Mi = 3 TO 5:
            
            IF TRIM(Mtoken[Mi]) = "ASCENDING" THEN
              Bifield._ascending = YES.
            
            IF TRIM(Mtoken[Mi]) = "DESCENDING" THEN
              Bifield._ascending = NO.
              
            IF TRIM(Mtoken[Mi]) = "ABBREVIATED" THEN
              Bifield._abbreviate = YES.
                 
            IF TRIM(Mtoken[Mi]) = "UNSORTED" THEN
              Bifield._unsorted = YES.
             
          END. /* Mi = 3 .. */
  
          Mtoken = "". 
          
        END. /* "INDEX-FIELD" */
        
        OTHERWISE DO:
          Perror = "Unknown token: " + (IF Mtoken[1] = ? THEN "?" ELSE TRIM(Mtoken[1])).
          LEAVE $TRANS-SCOPE.
        END.
        
        END CASE.
   
      END. /* INDEX-UPDATE */
  
      Bfile._frozen = Mold-frozen.
  
      Mtoken = "".
      IF Bfile._prime-index <> -999999 THEN
        NEXT $TRANS-SCOPE.
      ELSE
        LEAVE $TRANS-SCOPE.
        
    END. /* Mobject = "INDEX" */
  
  END. /* $READ-DDL: to read ddl file */

END. /* $TRANS-SCOPE: scopping of update */

IF Perror <> "" THEN DO:
  DISPLAY
    Mline
    Perror @ Mdisp
    WITH FRAME monitor.
  PAUSE.
END.
ELSE
  DISPLAY
    Mline
    "Load of schema complete" @ Mdisp
    WITH FRAME monitor.

INPUT STREAM Sin1 CLOSE.
